#!/usr/bin/env python

import os
import re

import sys

import json
import glob

import argparse
import subprocess

from deepsearch_glm.utils.load_pretrained_models import load_pretrained_nlp_models

ROOT_DIR=os.path.abspath("./")
BUILD_DIR=os.path.join(ROOT_DIR, "build")

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'apply_nlp_on_doc',
        description = 'Apply NLP on `Deep Search` documents',
        epilog =
"""
""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--mode', required=False,
                        type=str, default="local",
                        help="build-mode")

    args = parser.parse_args()

    return args.mode

def download_nlp_models():
    load_pretrained_nlp_models(False)

def run(cmd, cwd="./"):

    print(f"\nlaunch: {cmd}")
    
    parts = cmd.split(" ")
    message = subprocess.run(parts, cwd=cwd)    

    if "returncode=0" in str(message):
        print(f" -> SUCCESS")
        return True

    print(f" -> ERROR with message: '{message}'\n")        
    return False
    
def build_local(setup_kwargs=None):

    print("python executable: ", sys.executable)
    
    if not os.path.exists(BUILD_DIR):
        #cmd = f"cmake -B {BUILD_DIR}"
        cmd = f"cmake -B {BUILD_DIR} -DPYTHON_EXECUTABLE={sys.executable}"
        run(cmd, cwd=ROOT_DIR)
    else:
        print(f"build directory detected: {BUILD_DIR}")
        
    #cmd = f"cmake --build {BUILD_DIR} --target install -j"
    cmd = f"cmake --build {BUILD_DIR} --target install "
    run(cmd, cwd=ROOT_DIR)    

def build_all_python_versions():

    candidates = glob.glob("/usr/local/bin/python3.*")
    candidates += glob.glob("/usr/bin/python3.*")

    print(f"all candidates: {candidates}")
    
    python_versions=[]
    for candidate in candidates:
        pyname = os.path.basename(candidate)
        if re.match("^python3.\d+$", pyname):
            python_versions.append(candidate)

    python_versions = sorted(python_versions)
    print(f"all found python-versions: {python_versions}")
    
    for pyv in python_versions:

        pyn = os.path.basename(pyv)        
        PYBUILD_DIR = os.path.join(ROOT_DIR, f"build-{pyn}")
        if os.path.exists(PYBUILD_DIR):
            print(f"rm {PYBUILD_DIR}")
            continue
        
        cmd = f"cmake -B {PYBUILD_DIR} -DPYTHON_EXECUTABLE={pyv}"
        run(cmd, cwd=ROOT_DIR)
            
        cmd = f"cmake --build {PYBUILD_DIR} --target install -j"
        run(cmd, cwd=ROOT_DIR)    
    
if "__main__"==__name__:

    #mode = parse_arguments()
    
    #load_pretrained_nlp_models(False)

    build_local()
    #build_all_python_versions()




import os

import json
import copy
import glob
import hashlib
import datetime
import subprocess

from tqdm import tqdm
from numerize.numerize import numerize
from dotenv import load_dotenv

import deepsearch as ds

from deepsearch.cps.client.components.elastic import ElasticDataCollectionSource
from deepsearch.cps.queries import DataQuery
from deepsearch.cps.client.components.queries import RunQueryError

def get_scratch_dir():
    
    load_dotenv()

    tmpdir = os.path.abspath(os.getenv("DEEPSEARCH_GLM_SCRATCH_DIR"))

    if not os.path.exists(tmpdir):
        os.mkdir(tmpdir)

    return tmpdir
    
def load_vars():

    load_dotenv()
    
    host = os.getenv("DEEPSEARCH_HOST")
    proj = os.getenv("DEEPSEARCH_PROJ")
    
    username = os.getenv("DEEPSEARCH_USERNAME")
    apikey = os.getenv("DEEPSEARCH_APIKEY")

    verify_ssl = os.getenv("DEEPSEARCH_VERIFYSSL")

    # avoid common mistakes ...
    if host=="https://deepsearch-experience.res.ibm.com/":
        verify_ssl=True
    elif host=="https://cps.foc-deepsearch.zurich.ibm.com/":
        verify_ssl=False
    
    return host, proj, username, apikey, verify_ssl

def get_ds_api():

    tdir = get_scratch_dir()
    
    host, proj, username, apikey, verify_ssl = load_vars()
    
    config_ = {
        "host": host,
        "auth": {
            "username": username,
            "api_key": apikey
        },
        "verify_ssl": verify_ssl
    }

    config_file = f"{tdir}/ds_config.json"
    with open(config_file, "w") as fw:
        fw.write(json.dumps(config_))
    
    config = ds.DeepSearchConfig.parse_file(config_file)    
    client = ds.CpsApiClient(config)
    
    api = ds.CpsApi(client)

    return api, proj

def process_zip_files(tdir):

    zipfiles = sorted(glob.glob(os.path.join(tdir, "*.zip")))
    #print(f"zips: ", len(zipfiles))

    for zipfile in zipfiles:
        cmd = ["unzip", zipfile, "-d", tdir]
        print(" ".join(cmd))

        subprocess.call(cmd)

    # clean up
    for i,zipfile in enumerate(zipfiles):
        print(i, "\t removing ", zipfile)
        subprocess.call(["rm", zipfile])        

    """
    cellsfiles = sorted(glob.glob(os.path.join(tdir, "*.cells")))
    for i,cellsfile in enumerate(cellsfiles):
        subprocess.call(["rm", cellsfile])            
    """
    
def convert_pdffiles(pdf_files, force=False):

    for i,pdf_file in enumerate(pdf_files):

        old_file = pdf_file
        """
        if os.path.exists(f"{old_file}"):
            print("exists ...")
        """
        
        subprocess.call(["ls", f"{old_file}"])

        if " " in pdf_file:
            new_file = pdf_file.replace(" ", "_")
            subprocess.call(["mv", f"{old_file}", f"{new_file}"])
            pdf_files[i] = new_file
    
    json_files=[]
    
    new_pdfs=[]
    for pdf_file in pdf_files:

        if force:
            new_pdfs.append(pdf_file)
        else:        
            json_file = pdf_file.replace(".pdf", ".json")
            if os.path.exists(json_file):
                json_files.append(json_file)
            else:
                new_pdfs.append(pdf_file)

    if len(new_pdfs)==0:
        return json_files
    else:
        print("found new pdf's: ", json.dumps(new_pdfs, indent=2))
        
    scratch_dir = get_scratch_dir()    
    
    old_pdfs = glob.glob(os.path.join(scratch_dir, "*.pdf"))
    for old_pdf in old_pdfs:
        subprocess.call(["rm", f"{old_pdf}"])
    
    for new_pdf in new_pdfs:
        subprocess.call(["cp", f"{new_pdf}", scratch_dir])

    ds_api, ds_proj = get_ds_api()
    
    documents = ds.convert_documents(api=ds_api, proj_key=ds_proj,
                                     source_path=scratch_dir,
                                     progress_bar=True)           

    documents.download_all(result_dir=scratch_dir)
    process_zip_files(scratch_dir)

    info = documents.generate_report(result_dir=scratch_dir)
    
    for new_pdf in new_pdfs:
        basename = os.path.basename(new_pdf).replace(".pdf", ".json")
        
        sfile = os.path.join(scratch_dir, basename)
        tfile = new_pdf.replace(".pdf", ".json")

        cmd = ["cp", f"{sfile}", f"{tfile}"]
        print(" ".join(cmd))

        subprocess.call(cmd)
        json_files.append(tfile)

    json_files = sorted(list(set(json_files)))
    return json_files

def ds_list_indices():

    api, proj_key = get_ds_api()
    
    # Fetch list of all data collections
    collections = api.elastic.list()
    collections.sort(key=lambda c: c.name.lower())    

    # Visualize summary table
    results = [
        {
            "Name": c.name,
            "Type": c.metadata.type,
            "Num entries": numerize(c.documents),
            "Date": c.metadata.created.strftime("%Y-%m-%d"),
            "Id": f"{c.source.elastic_id}",
            "Index": f"{c.source.index_key}",
        }
        for c in collections
    ]
    return results

def create_docs_dir():

    tdir = get_scratch_dir()

    now = datetime.datetime.now()
    odir = now.strftime("document-collection-%Y-%m-%d_%H-%M-%S")

    odir = os.path.join(tdir, odir)
    return odir    

def ds_index_query(index, query, odir=None, force=False, limit=-1,
                   sources=["file-info", "description",
                            "main-text",
                            "texts", "tables", "figures",                
                            "page-headers", "page-footers",
                            "footnotes"] # Which fields of documents we want to fetch
                   ):
    
    api, proj_key = get_ds_api()
    
    tdir = get_scratch_dir()

    if odir==None:
        dumpdir = hashlib.md5(f"{index}:{query}".encode()).hexdigest()
        dumpdir = os.path.join(tdir, dirname)
    else:
        dumpdir = odir
        
    if not os.path.exists(dumpdir):
        os.mkdir(dumpdir)
    elif not force:
        return dumpdir
    
    # Input query
    search_query = f"{query}" #"\"global warming potential\" AND \"etching\""
    print(f"query: {query}")
    
    data_collection = ElasticDataCollectionSource(elastic_id="default", index_key=index)
    page_size = 50

    """
        source=["file-info", "description",
                "main-text",
                "texts", "tables", "figures",                
                "page-headers", "page-footers",
                "footnotes"], 
    """
    
    # Prepare the data query
    query = DataQuery(
        search_query, # The search query to be executed
        source=sources, # Which fields of documents we want to fetch
        limit=page_size, # The size of each request page
        coordinates=data_collection # The data collection to be queries
    )

    # [Optional] Compute the number of total results matched. This can be used to monitor the pagination progress.
    count_query = copy.deepcopy(query)
    count_query.paginated_task.parameters["limit"] = 0
    count_results = api.queries.run(count_query)
       
    expected_total = count_results.outputs["data_count"]
    print(f"#-found documents: {expected_total}")

    if limit!=-1 and expected_total>limit:
        expected_total=limit    

    expected_pages = (expected_total + page_size - 1) // page_size # this is simply a ceiling formula
    
    # Iterate through all results by fetching `page_size` results at the same time
    bar_format = '{desc:<5.5}{percentage:3.0f}%|{bar:70}{r_bar}'

    count=0
    
    all_results = []
    cursor = api.queries.run_paginated_query(query)
    for result_page in tqdm(cursor, total=expected_pages, bar_format=bar_format):
        for row in result_page.outputs["data_outputs"]:
            _id = row["_id"]
            with open(f"{dumpdir}/{_id}.json", "w") as fw:
                fw.write(json.dumps(row["_source"], indent=2))

            count+=1
            if limit!=-1 and count>=limit:
                break

        if limit!=-1 and count>=limit:
            break            
                
    return dumpdir




#!/usr/bin/env python

import os

import json
import glob

import argparse
import textwrap
import datetime

from tabulate import tabulate

from deepsearch_glm.utils.ds_utils import get_scratch_dir

from deepsearch_glm.andromeda_glm import glm_model

#import andromeda_nlp
#import andromeda_glm

def create_glm_dir():

    tdir = get_scratch_dir()

    now = datetime.datetime.now()
    glmdir = now.strftime("GLM-model-%Y-%m-%d_%H-%M-%S")

    odir = os.path.join(tdir, glmdir)
    return odir

def load_glm_config(idir:str):

    config = {
        "IO": {
            "load": {
                "root": idir
            },
        },
    }

    return config

def load_glm(idir:str):

    config = load_glm_config(idir)
    
    #glm = andromeda_glm.glm_model()
    glm = glm_model()
    glm.load(config)

    return glm

def create_glm_config_from_docs(odir:str, json_files:list[str],
                                nlp_models:str="conn;verb;term;abbreviation"):

    config = {
        "IO": {
            "load": {
                "root": odir
            },
            "save": {
                "root": odir,
                "write-CSV": True,
                "write-JSON": False,
                "write-path-text": False
            }
        },
        "create": {
            "enforce-max-size": False,
            "model": {
                "max-edges": 1e8,
                "max-nodes": 1e7
            },
            "number-of-threads": 4,
            "worker": {
                "local-reading-break": True,
                "local-reading-range": [
                    256,
                    2560
                ],
                "max-edges": 1e7,
                "max-nodes": 1e6
            },
            "write-nlp-output": False
        },
        "mode": "create",
        "parameters": {
            "glm-padding": 1,
            "glm-paths": {
                "keep-concatenation": True,
                "keep-connectors": True,

                "keep-terms": True,
                "keep-verbs": True,
                
                "keep-sentences": True,
                
                "keep-tables": True,
                "keep-texts": True,

                "keep-docs": True,
            },

            "nlp-models": "conn;verb;term;abbreviation"
        },
        "producers": [
            {
                "input-format": "json",
                "input-paths": json_files,
                
                "keep-figures": True,
                "keep-tables": True,
                "keep-text": True,

                "order-text": True,

                "output": False,
                "output-format": "nlp.json",

                "subject-type": "DOCUMENT"
            }
        ]
    }

    return config

def create_glm_from_docs(odir:str, json_files:list[str],
                         nlp_models:str="conn;verb;term;abbreviation"):
    
    config = create_glm_config_from_docs(odir, json_files, nlp_models)
    
    #glm = andromeda_glm.glm_model()
    glm = glm_model()
    glm.create(config)

    return odir, glm


    



#!/usr/bin/env python

import os

import json
import glob

import argparse
import textwrap
import datetime

from tabulate import tabulate

from deepsearch_glm.glm_utils import load_glm

from deepsearch_glm.andromeda_nlp import nlp_model
from deepsearch_glm.andromeda_glm import glm_model, glm_query

#import andromeda_nlp
#import andromeda_glm

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'explore_glm',
        description = 'Explore GLM from Deep Search documents',
        epilog =
"""
examples of execution: 

1 Explore the GLM
    poetry run python ./deepsearch_glm/explore_glm.py --glm <glm-root-dir>

""",
        formatter_class=argparse.RawTextHelpFormatter)
        
    parser.add_argument('--glm', required=True,
                        type=str,  help="root directory of GLM")
    
    args = parser.parse_args()

    if not os.path.exists(args.glm):
        exit(-1)
        
    return args.glm

def show_query_result(res, max_nodes=16):

    wrapper = textwrap.TextWrapper(width=50)
    
    print("overview: \n", tabulate(res["overview"]["data"],
                                   headers=res["overview"]["headers"]), "\n")

    for i,item in enumerate(res["result"]):
        
        headers = item["nodes"]["headers"]
        data = item["nodes"]["data"]

        for j,row in enumerate(data):
            text = row[headers.index("text")]
            #print("text: ", text)
            
            data[j][headers.index("text")] = "\n".join(wrapper.wrap(text))
                    
        print(f"operation {i}: \n", tabulate(data[0:max_nodes], headers=headers), "\n")

def expand_terms(terms, glm):

    for term in terms:
        print(term)
        
        #qry = andromeda_glm.glm_query()
        qry = glm_query()
        
        qry.select({"nodes":[[term]]})
        qry.filter_by({"mode": "node-flavor", "node-flavors":["token"]})
        #qry.filter_by({"mode": "node-flavor", "node-flavors":["term"]})

        #flid = qry.get_last_flid()
        qry.traverse({"name": "roots", "edge":"to-root"})
        qry.traverse({"name": "tax-up", "edge":"tax-up"})
        #qry.traverse({"edge":"from-root"})
        #qry.traverse({"edge":"from-token"})
        #qry.filter_by({"mode": "node-flavor", "node-flavors":["term"]})

        #qry.filter_by({"mode": "contains", "contains-flid":flid})        
        #qry.traverse({"edge":"to-sent"})

        qry.filter_by({"mode": "node-regex", "node-regex":[f".*{term}.*"]})
        
        config = qry.to_config()    
        #print("query: ", json.dumps(config, indent=2))    
        
        res = glm.query(config)
        if "status" in res and res["status"]=="success":
            show_query_result(res)
        else:
            print(res)
            #print(res["status"], ": ", res["message"])

def explore(glm_model):
    
    while True:
        
        prompt = input("question: ")
        #prompt = "What is the income of IBM in 2022?"
        #prompt = "net-zero"
        if(prompt=="q"):
            break
        
        #terms = analyse_prompt(prompt, nlp_model)
        terms = prompt.split(" ")
        
        expand_terms(terms, glm_model)            

if __name__ == '__main__':

    idir = parse_arguments()

    glm = load_glm(idir)
    
    explore(glm)




import os
import json
import subprocess

def load_pretrained_nlp_models(force:bool=False, verbose:bool=False):

    if "DEEPSEARCH_GLM_RESOURCES_DIR" in os.environ:
        RESOURCES_DIR = os.getenv("DEEPSEARCH_GLM_RESOURCES_DIR")
    else:
        from deepsearch_glm.andromeda_nlp import nlp_model

        model = nlp_model()
        RESOURCES_DIR = model.get_resources_path()
    
    with open(f"{RESOURCES_DIR}/models.json") as fr:
        models = json.load(fr)
        
    COS_URL = models["object-store"]
    COS_PRFX = models["nlp"]["prefix"]
    COS_PATH = os.path.join(COS_URL, COS_PRFX)
    
    cmds=[]
    for name,files in models["nlp"]["trained-models"].items():
        source = os.path.join(COS_PATH, files[0])
        target = os.path.join(RESOURCES_DIR, files[1])

        cmd = ["curl", source, "-o", target, "-s"]
        cmds.append(cmd)
        
    for cmd in cmds:

        model_weights = cmd[3]
        
        if force or (not os.path.exists(model_weights)):

            if verbose:
                print(f"downloading {os.path.basename(model_weights)} ... ", end="")            

            message = subprocess.run(cmd)#, cwd=ROOT_DIR)    

            if verbose:
                print("done!")
                
        elif verbose:
            print(f" -> already downloaded {os.path.basename(cmd[3])}")



#!/usr/bin/env python

import os
import re

import time
import json
import glob
import argparse

import random

import subprocess

import numpy as np

import pandas as pd
import matplotlib.pyplot as plt

#import fasttext
import textColor as tc

#import deepsearch as ds
#from tabulate import tabulate

#import andromeda_nlp
from deepsearch_glm.andromeda_nlp import nlp_model

from deepsearch_glm.utils.ds_utils import convert_pdffiles
from deepsearch_glm.nlp_utils import create_nlp_dir

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = "nlp_train_reference",
        description = 'Prepare CRF data for CRF-reference parser',
        epilog =
"""
examples of execution: 

1. end-to-end example on pdf documents:

    poetry run python ./deepsearch_glm/nlp_train_reference.py -m all --pdf './data/documents/articles/*.pdf'

""",
        formatter_class=argparse.RawTextHelpFormatter)
        
    parser.add_argument('-m', '--mode', required=False, default="all",
                        help="parse: [convert,extract,annotate,classify,pure-classify,crf,pure-crf,all]")

    parser.add_argument('--pdf', required=True,
                        type=str, default=None,
                        help="filename(s) of pdf document")

    parser.add_argument('--json', required=False,
                        type=str, default=None,
                        help="filename(s) of json document")
    
    parser.add_argument('--output-dir', required=False,
                        type=str, default=create_nlp_dir(),
                        help="output root directory for trained models")
    
    """
    parser.add_argument('', '--source-directory', required=False, default="./data/documents/articles",
                        help="directory with pdfs")
    parser.add_argument('-t', '--target-directory', required=False, default="./data/models/",
                        help="directory for target files")
    """
    
    args = parser.parse_args()

    pdf = args.pdf
    json = args.json

    if pdf==None and json==None:
        exit(-1)
        
    if pdf!=None:
        pdf_files=sorted(glob.glob(pdf))
    else:
        pdf_files=[]
        
    if json!=None:
        json_files=sorted(glob.glob(json))
    else:
        json_files=[]
    
    if not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
    
    return args.mode, pdf_files, json_files, args.output_dir

def shorten_text(text):
    
    ntext = text.replace("\n", "")
    
    return ntext.strip()
        
def extract_references(filenames, sfile, rfile):

    print(f"extract references for filenames: ", len(filenames))
    
    config = {
        "mode" : "apply",
        "order" : True,
        "models": "numval,link"
    }
    
    #model = andromeda_nlp.nlp_model()
    model = nlp_model()
    model.initialise(config)
    
    MINLEN = 5

    fws = open(sfile, "w")
    fwr = open(rfile, "w")
    
    for filename in filenames:

        if filename.endswith("references.json"):
            continue
        
        print(f"reading {filename}")

        try:
            with open(filename, "r") as fr:
                data = json.load(fr)
        except:
            continue
        
        with open(filename, "w") as fw:
            fw.write(json.dumps(data, indent=2))             

        is_ref=False
        cnt_ref=0
        
        for item in data["main-text"]:

            if "text" in item:
                text = item["text"].strip()
            else:
                continue

            if "type" in item:
                label = item["type"]
            else:
                continue

            label = (label.split("-"))[0]            

            content = text.lower().strip().replace(" ", "")
            
            if content.endswith("references"):
                is_ref = True
            elif is_ref and label=="subtitle":
                is_ref = False

            if is_ref and (not content.endswith("references")) and len(text)>=MINLEN:                
                label = "reference"
            elif re.match("^(\d+|\[\d+\])(.*)\((19|20)\d{2}\)\.?$", text):
                label = "reference"
            elif re.match("^(\[\d+\])(\s+[A-Z]\.)+.*", text):
                label = "reference"                
            elif re.match("^(\[\])(.*)\((19|20)\d{2}\)\.?$", text):
                label = "reference"
            elif re.match("^(Table|Figure)(\s+\d+(\.\d+)?)(.*)", text):
                label = "caption"                
            elif len(text.strip())<MINLEN:
                label = "paragraph"
            elif label=="title":
                label = "subtitle"

            nlpres = model.apply_on_text(text)
            nlpres["filename"] = os.path.basename(filename.replace(".json", ".pdf"))

            if(label=="subtitle"):            
                print(nlpres["filename"], "\t", label, "\t", tc.yellow(text[0:96]))                
            elif(label=="reference"):            
                print(nlpres["filename"], "\t", label, "\t\t", tc.blue(text[0:96]))
            elif(label=="caption"):            
                print(nlpres["filename"], "\t", label, "\t\t", tc.red(text[0:96]))                                
            else:
                print(nlpres["filename"], "\t", label, "\t\t", text[0:96])
                
            if label=="reference":
                cnt_ref += 1
                fwr.write(json.dumps(nlpres)+"\n")

            if True:

                """
                print(tabulate(nlpres["word-tokens"]["data"],
                               headers=nlpres["word-tokens"]["headers"]))
                """

                """
                print(tabulate(nlpres["properties"]["data"],
                               headers=nlpres["properties"]["headers"]))
                """
                
                tind = nlpres["properties"]["headers"].index("type")
                lind = nlpres["properties"]["headers"].index("label")
                cind = nlpres["properties"]["headers"].index("confidence")

                found=False
                for i,row in enumerate(nlpres["properties"]["data"]):
                    if row[tind]=="semantic":
                        row[lind] = label
                        row[cind] = 1.0

                        found = True

                if not found:
                    nlpres["properties"]["data"].append(["semantic", label, 1.0])
                    found = True

                if random.random()<0.9:
                    nlpres["training-sample"] = True
                else:
                    nlpres["training-sample"] = False
                    
                if found:                    
                    fws.write(json.dumps(nlpres)+"\n")
                
        if cnt_ref>0:
            print(tc.green(f"{filename}: {cnt_ref}"))
        else:
            print(tc.yellow(f"{filename}: {cnt_ref}"))
                
    fws.close()
    fwr.close()

    print(f"semantic-classification dumped in {sfile}")
    print(f"references dumped in {rfile}")

def parse_with_anystyle_api(tlines):

    time.sleep(1)

    tmpfile = "tmp.json"
    
    payload = { "input": [] }
    for tline in tlines:
        payload["input"].append(tline[1])

    anystyle_token = '9fEhg+39p0J60Bs+WTTwTMcqqTFAUYoyjLlp8nEys4wnfgACn0IoqravX8Exsx/+2q1p4sU7636DR22xUeneLg=='
    anystyle_session = '9GFKMlFoJwbMV6W1Z37YFsG9nbXLqmGicXVzL4r5mn4SqTLcf0revMMFvAjfxcjqR8YBnj2M0fgTWBW12kK1KMFcOgZvZnwQv5lZZ3PQgPP9sait9WgoDR72BHqRpbPe0c1B6%2BNFtYE7aqpugLsTupqBuj%2B%2Fef0tbyd84wC61GkVA9Vtz2nSNC90hDliCre%2BZ2gQUc6runu6yt1M4xa0F8kM4Cxt2pN92XB8hRusqGNfsaCsw5JKdU%2FcDFtdh%2BYDSEBz6DjQFfJq81%2FTI%2F4ulku7mlv73vOC7ew%3D--o%2B2gjgNJqgCjYf4V--3mSN%2FKmNt68WTJsxBh9Bww%3D%3D'
        
    res = subprocess.call(['curl', 'https://anystyle.io/parse?format=json',
                           '-H', 'authority: anystyle.io'  ,
                           '-H', 'accept: application/json, text/plain, */*'  ,
                           '-H', 'accept-language: en-GB,en-US;q=0.9,en;q=0.8'  ,
                           '-H', 'content-type: application/json;charset=UTF-8'  ,
                           '-H', f'cookie: _any_style_session={anystyle_session}',
                           '-H', 'origin: https://anystyle.io'  ,
                           '-H', 'referer: https://anystyle.io/'  ,
                           '-H', 'sec-ch-ua: "Not_A Brand";v="99", "Google Chrome";v="109", "Chromium";v="109"'  ,
                           '-H', 'sec-ch-ua-mobile: ?0'  ,
                           '-H', 'sec-ch-ua-platform: "macOS"'  ,
                           '-H', 'sec-fetch-dest: empty'  ,
                           '-H', 'sec-fetch-mode: cors'  ,
                           '-H', 'sec-fetch-site: same-origin'  ,
                           '-H', 'user-agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36'  ,
                           '-H', f'x-csrf-token: {anystyle_token}',
                           '--data-raw', json.dumps(payload),
                           '--compressed', '-o', tmpfile, '-s'])

    try:
        with open(tmpfile, "r") as fr:
            tmp = json.load(fr)

        if os.path.exists(tmpfile):
            os.remove(tmpfile)
            
        return tmp
    except:
        print(tc.red("could not call anystyle API endpoint ..."))

    if os.path.exists(tmpfile):
        os.remove(tmpfile)
        
    return []

def update_references(refs, tlines):

    results = parse_with_anystyle_api(tlines)
    
    for j,item in enumerate(results):
                
        parts=[]
        for row in item:
            parts.append(row[1])
            
        text = " ".join(parts)
        if text!=tlines[j][1]:
            continue

        beg=0
        for k,row in enumerate(item):
            charlen = len(row[1].encode('utf-8'))
            
            item[k].append(beg)
            item[k].append(beg+charlen)

            beg += charlen
            beg += 1

        ind = tlines[j][0]
        
        refs[ind]["word-tokens"]["headers"].append("true-label")
                
        for ri,row_i in enumerate(refs[ind]["word-tokens"]["data"]):

            label="__undef__"
            for rj,row_j in enumerate(item):
                if row_j[2]<=row_i[0] and row_i[1]<=row_j[3]:
                    label = row_j[0]
                    break

            refs[ind]["word-tokens"]["data"][ri].append(label)

        """
        print(tabulate(refs[ind]["word-tokens"]["data"],
                       headers=refs[ind]["word-tokens"]["headers"]))
        """
        
        refs[ind]["annotated"]=True
            
    tlines=[]

def annotate(rfile, ofile):

    refs=[]

    fr = open(rfile, "r")

    while True:

        line = fr.readline().strip()
        if line==None or len(line)==0:
            break

        try:
            item = json.loads(line)
            refs.append(item)
        except:
            continue

    fr.close()

    print("#-refs: ", len(refs))
    
    tlines=[]
    for ind,ref in enumerate(refs):

        print(f"\rreferennce-annotation: {ind}/{len(refs)}", end="")
        
        refs[ind]["annotated"]=False
        tlines.append([ind, ref["text"]])
        
        if len(tlines)>0 and len(tlines)%16==0:
            update_references(refs, tlines)
            tlines=[]

    print(" --> done")
            
    if len(tlines)>0:
        update_references(refs, tlines)
            
    fw = open(ofile, "w")    

    for ref in refs:
        if "annotated" in ref and ref["annotated"]:
            fw.write(json.dumps(ref)+"\n")

    fw.close()
        
    print(f"writing annotation to {ofile}")

def prepare_for_crf(afile):

    labels={}
    
    refs=[]
    
    fr = open(afile, "r")

    while True:

        line = fr.readline().strip()
        #print(line)
        
        if line==None or len(line)==0:
            break

        try:
            item = json.loads(line)            
        except:
            continue

        wt = item["word-tokens"]

        if item["annotated"]:

            tind = wt["headers"].index("word")
            lind = wt["headers"].index("true-label")

            for i,row in enumerate(wt["data"]):

                if row[tind] in ["(", "[", "{"]:
                    row[lind] = "obracket"

                if row[tind] in [")", "]", "}"]:
                    row[lind] = "cbracket"

                if row[tind] in [",", ".", "'", "\"", ";", ":"]:                    
                    row[lind] = row[tind]

                if row[tind] in ["and", "&"] and row[lind]=="author":                    
                    row[lind] = "author_and"                                        

                if row[tind] in ["et", "al"] and row[lind]=="author":                    
                    row[lind] = "author_etal"                                        

            tokens = []
            for i,row in enumerate(wt["data"]):

                text = wt["data"][i][tind]
                text = text.replace(" ", "")

                wt["data"][i][tind] = text
                tokens.append(wt["data"][i][tind])
                
            line = " ".join(tokens)

            if True: # update true-labels for authors 
                for i,row in enumerate(wt["data"]):
                
                    if i>0 and \
                       wt["data"][i-1][lind]=="author" and \
                       wt["data"][i][lind] in ["."]:
                        wt["data"][i][lind] = "author"

                ## example:
                ##  - `Wright, N. J., Drake, J. J., Mamajek, E. E., & Henry, G. W. 2011, The Astrophysical Journal, 743, 48, doi: 10.1088/0004-637x/743/1/48`,
                ## `- 5 . Bender , C . M . ; Orszag , S . A . : Advanced Mathematical Methods for Scientists and Engineers : Asymptotic Methods and Perturbation Theory Springer - Verlag New - York Inc , 1999`
                if((" . , " in line) or (" . ; " in line) or (" . : " in line)):

                    for i,row in enumerate(wt["data"]):
                        
                        if i>0 and i<len(wt["data"])-1 and \
                           wt["data"][i-1][tind]!="." and \
                               wt["data"][i-1][lind]=="author" and \
                               wt["data"][i+1][lind]=="author" and \
                               wt["data"][i][lind] in [","]:
                            wt["data"][i][lind] = "author"

            if True: # update true-labels for title, journal, editor, container-title, location, date
                
                for i,row in enumerate(wt["data"]):

                    if 0<i and i<len(wt["data"])-1 and \
                       wt["data"][i-1][lind] in ["title", "journal", "editor", "container-title", "volume", "pages", "location", "date", "url"] and \
                       wt["data"][i-1][lind]==wt["data"][i+1][lind] and \
                       len(wt["data"][i][lind])==1:
                        wt["data"][i][lind] = wt["data"][i-1][lind]

                for i,row in enumerate(wt["data"]):

                    if 0<i and i<len(wt["data"])-1 and \
                       wt["data"][i-1][lind] in ["title", "volume", "pages"] and \
                       wt["data"][i-0][lind]== "cbracket":
                        wt["data"][i][lind] = wt["data"][i-1][lind]                        
                    
            for i,row in enumerate(wt["data"]):

                if " " in wt["data"][i][tind]:
                    has_space=True
                
                label = wt["data"][i][lind]
                if label in labels:
                    labels[label] += 1
                else:
                    labels[label] = 1
            
            if item["filename"]=="2302.05256.pdf":
                print("text: ", item["text"])
                print(tabulate(wt["data"], headers=wt["headers"]), "\n")

            refs.append(item)

    fr.close()

    print(json.dumps(labels, indent=2))
    
    doc_to_inds={}
    
    for i,ref in enumerate(refs):

        refs[i]["training-sample"] = True

        fname = ref["filename"]
        if fname in doc_to_inds:
            doc_to_inds[fname].append(i)
        else:
            doc_to_inds[fname] = [i]

    # pick 10% random reference from each paper for validation
    for key,val in doc_to_inds.items():

        num = max(1, int(0.1*len(val)))
        print(key, " -> ", num, " / ", len(val))
        
        for l in range(0, num):
            k = int(random.random()*len(val))
            refs[val[k]]["training-sample"] = False

    fw = open(afile, "w")
    for i,ref in enumerate(refs):
        fw.write(json.dumps(ref)+"\n")
    fw.close()

    print(tc.green(f"training file has been written to {afile}"))

def train_crf(train_file, model_file, metrics_file):

    #model = andromeda_nlp.nlp_model()
    model = nlp_model()
                
    configs = model.get_train_configs()

    for config in configs:
        if config["mode"]=="train" and \
           config["model"]=="reference":

            config["files"]["model-file"] = model_file
            config["files"]["train-file"] = train_file
            config["files"]["metrics-file"] = metrics_file 
    
            model.train(config)


# To train a FST model with HPO, one can use
# 
# `./fasttext supervised -input <path-to-train.txt> -output model_name -autotune-validation <<path-to-valid.txt>> -autotune-duration 600 -autotune-modelsize 1M`
#
#  => the parameters can be found via
#
# `./fasttext dump model_cooking.bin args`
#
def train_fst(train_file, model_file, metrics_file):

    #model = andromeda_nlp.nlp_model()
    model = nlp_model()
                
    configs = model.get_train_configs()    
    print(configs)
    
    for config in configs:
        if config["mode"]=="train" and \
           config["model"]=="semantic":

            config["files"]["model-file"] = model_file
            config["files"]["train-file"] = train_file 
            config["files"]["metrics-file"] = metrics_file
            
            model.train(config)
            
if __name__ == '__main__':

    mode, pdf_files, json_files, tdir = parse_arguments()
    
    if len(pdf_files)>0:
        new_json_files = convert_pdffiles(pdf_files, force=False)

        for _ in new_json_files:
            json_files.append(_)

    json_files = sorted(list(set(json_files)))        
    
    sfile = os.path.join(tdir, "nlp-train-semantic-classification.annot.jsonl")
    rfile = os.path.join(tdir, "nlp-train-references-crf.jsonl")
    afile = os.path.join(tdir, "nlp-train-references-crf.annot.jsonl")
    
    crf_model_file = os.path.join(tdir, "crf_reference")
    fst_model_file = os.path.join(tdir, "fst_sematic")

    if mode=="extract" or mode=="all":
        extract_references(json_files, sfile, rfile)

    if mode=="annotate" or mode=="all":
        annotate(rfile, afile)

    if "classify" in mode or mode=="all":

        if mode=="classify" or mode==all:
            extract_references(sdir, sfile, rfile)
            
        train_fst(sfile, fst_model_file, fst_model_file+".metrics.txt")

    if "crf" in mode or mode=="all":

        if mode=="crf" or mode=="all":
            prepare_for_crf(afile)
        
        train_crf(afile, crf_model_file, crf_model_file+".metrics.txt")
        



#!/usr/bin/env python

import os
import json
import glob
import argparse

import pandas as pd

from utils.ds_utils import convert_pdffiles

#import andromeda_nlp
from deepsearch_glm.andromeda_nlp import nlp_model

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'apply_nlp_on_doc',
        description = 'Apply NLP on `Deep Search` documents',
        epilog =
"""
examples of execution: 

1.a run on single document (pdf or json) with default model (=`langauge`):

     poetry run python ./deepsearch_glm/apply_nlp_on_docs.py --pdf './data/documents/articles/2305.02334.pdf'
     poetry run python ./deepsearch_glm/apply_nlp_on_docs.py --json './data/documents/articles/2305.02334.json'

1.b run on single document pdf document and enforce conversion (ignore cache):

     poetry run python ./deepsearch_glm/apply_nlp_on_docs.py --pdf './data/documents/articles/2305.02334.pdf' --force-convert True

2. run on multiple documents:

     poetry run python ./deepsearch_glm/apply_nlp_on_docs.py --pdf './data/documents/articles/*.pdf'
     poetry run python ./deepsearch_glm/apply_nlp_on_docs.py --json './data/documents/articles/*.json'

3. run on multiple documents with non-default models:

     poetry run python ./deepsearch_glm/apply_nlp_on_docs.py --pdf './data/documents/articles/2305.*.pdf' --models 'language;term'

""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--pdf', required=False,
                        type=str, default=None,
                        help="filename(s) of pdf document")
    
    parser.add_argument('--json', required=False,
                        type=str, default=None,
                        help="filename(s) of json document")

    parser.add_argument('--models', required=False,                        
                        type=str, default="language;semantic",
                        help="set NLP models (e.g. `language` or `term;sentence`)")

    parser.add_argument('--force-convert', required=False,
                        type=bool, default=False,
                        help="force pdf conversion")
    
    args = parser.parse_args()

    pdf = args.pdf
    json = args.json

    if pdf==None and json==None:
        exit(-1)
        
    if pdf!=None:
        pdf_files=sorted(glob.glob(pdf))
    else:
        pdf_files=[]
        
    if json!=None:
        json_files=sorted(glob.glob(json))
    else:
        json_files=[]
        
    return pdf_files, json_files, args.models, args.force_convert

# FIXME: to be replaced with function in nlp_utils
def init_nlp_model(models:str, filters:list[str]=[]):
    
    #model = andromeda_nlp.nlp_model()
    model = nlp_model()
    
    config = model.get_apply_configs()[0]

    config["models"] = models
    config["subject-filters"] = filters
    
    model.initialise(config)

    return model

def show_doc(doc_j):

    """
    print('page-elements')
    print(json.dumps(doc_j["page-elements"][0:10], indent=2))
    
    print('main-text')
    print(json.dumps(doc_j["main-text"][0:10], indent=2))

    print('body')
    print(json.dumps(doc_j["body"][0:10], indent=2))

    print('meta')
    print(json.dumps(doc_j["meta"][0:10], indent=2))        
        
    print('texts')
    print(json.dumps(doc_j["texts"][0:10], indent=2))
    
    print('figures')
    print(json.dumps(doc_j["figures"][0], indent=2))

    print('tables')
    print(json.dumps(doc_j["tables"][0], indent=2))
    """        
    
    props = pd.DataFrame(doc_j["properties"]["data"],
                         columns=doc_j["properties"]["headers"])
    print("properties: \n\n", props)

    inst = pd.DataFrame(doc_j["instances"]["data"], 
                        columns=doc_j["instances"]["headers"])
    print("instances: \n\n", inst)

    terms = inst[inst["type"]=="term"]
    print("terms: \n\n", terms)

    hist = terms["hash"].value_counts()
    for key,val in hist.items():
        name = terms[terms["hash"]==key].iloc[0]["name"]
        print(f"{val}\t{name}")
    
    """
    ents = pd.DataFrame(doc_j["entities"]["data"], 
                        columns=doc_j["entities"]["headers"])
    print(ents)
    """
    
if __name__ == '__main__':

    pdf_files, json_files, model_names, force_convert = parse_arguments()

    if len(pdf_files)>0:
        new_json_files = convert_pdffiles(pdf_files, force=force_convert)

        for _ in new_json_files:
            json_files.append(_)

    json_files = sorted(list(set(json_files)))        
    
    #filters = []
    filters = ["properties", "instances"]
    
    model = init_nlp_model(model_names, filters)

    for json_file in json_files:

        print(f"reading {json_file} ... ", end="")
        with open(json_file, "r") as fr:
            doc_i = json.load(fr)

        print(f"applying models ... ", end="")
        doc_j = model.apply_on_doc(doc_i)

        print(doc_j.keys())
        show_doc(doc_j)
        
        nlp_file = json_file.replace(".json", ".nlp.json")
        print(f"writing  models {nlp_file}")
        
        with open(nlp_file, "w") as fw:
            fw.write(json.dumps(doc_j, indent=2))


    




import os

import json
import glob

import argparse
import textwrap

import datetime
import subprocess

import textColor as tc
import pandas as pd

from tabulate import tabulate

from deepsearch_glm.utils.ds_utils import get_scratch_dir

from deepsearch_glm.andromeda_nlp import nlp_model
#import andromeda_nlp

def create_nlp_dir(tdir=None):

    if tdir==None:
        tdir = get_scratch_dir()

    now = datetime.datetime.now()
    nlpdir = now.strftime("NLP-model-%Y-%m-%d_%H-%M-%S")

    odir = os.path.join(tdir, nlpdir)
    return odir

def list_nlp_model_configs():

    configs = []

    configs += nlp_model.get_apply_configs()
    configs += nlp_model.get_train_configs()

    return configs

def init_nlp_model(model_names:str="language;term"):

    #model = andromeda_nlp.nlp_model()
    model = nlp_model()

    configs = model.get_apply_configs()
    #print(json.dumps(configs, indent=2))
    
    config = model.get_apply_configs()[0]
    config["models"] = model_names

    model.initialise(config)
    
    return model

def print_key_on_shell(key, items):

    df = pd.DataFrame(items["data"],
                      columns=items["headers"])
    
    if key in ["instances", "entities"]:

        wrapper = textwrap.TextWrapper(width=70)
        
        df = df[["type", "subtype", "subj_path", "char_i", "char_j", "original"]]

        table=[]
        for i,row in df.iterrows():
            _=[]
            for __ in row:
                if isinstance(__,str): 
                    _.append("\n".join(wrapper.wrap(__)))
                else:
                    _.append(__)

            table.append(_)

        headers = ["type", "subtype", "subj_path", "char_i", "char_j", "original"]
        print(tc.yellow(f"{key}: \n\n"), tabulate(table, headers=headers), "\n")
                         
    else:
        df = pd.DataFrame(items["data"],
                          columns=items["headers"])
        
        print(tc.yellow(f"{key}: \n\n"), df.to_string(), "\n")

def print_on_shell(text, result):

    wrapper = textwrap.TextWrapper(width=70)    
    print(tc.yellow(f"\ntext: \n\n"), "\n".join(wrapper.wrap(text)), "\n")
    
    for _ in ["properties", "word-tokens", "instances",
              "entities", "relations"]:
        if _ in result and len(result[_]["data"])>0:
            print_key_on_shell(_, result[_])
        else:
            print(tc.yellow(f"{_}:"), " null\n\n")
    



#!/usr/bin/env python

import os
import re

import time
import json
import glob

import random
import argparse

import textColor as tc

from deepsearch_glm.andromeda_nlp import nlp_model

from deepsearch_glm.utils.ds_utils import convert_pdffiles, ds_index_query
from deepsearch_glm.nlp_utils import create_nlp_dir, init_nlp_model, print_on_shell

#import andromeda_nlp

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = "nlp_train_semantic",
        description = 'train classifier for semantic text classifier',
        epilog =
"""
examples of execution: 

1. end-to-end example on pdf documents:

    poetry run python ./deepsearch_glm/nlp_train_semantic.py -m all --json '<root-dir-of-json-docs>'

""",
        formatter_class=argparse.RawTextHelpFormatter)
        
    parser.add_argument('-m', '--mode', required=False, default="all",
                        help="parse: [retrieve,prepare,train,eval,all]")

    parser.add_argument('--output-dir', required=False,
                        type=str, default=None,
                        help="output root directory for trained models")
    
    args = parser.parse_args()

    """
    if args.json!=None:        
        json_files=sorted(glob.glob(args.json))    
    else:
        json_files=None
    """
    if args.output_dir==None:
        odir = create_nlp_dir()
    
    elif not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
        odir = args.output_dir

    else:
        odir = args.output_dir
        
    return args.mode, odir

def retrieve_data_pubmed(sdir):

    tdir = os.path.join(sdir, "pubmed")

    if not os.path.exists(sdir):
        os.mkdir(sdir)
    
    index="pubmed"
    query="*"
        
    odir = ds_index_query(index, query, tdir, sources=["_name", "file-info", "references", "description"],
                          force=True, limit=1000)        
    
    return odir

def retrieve_data(sdir, index):

    tdir = os.path.join(sdir, index)

    if not os.path.exists(sdir):
        os.mkdir(sdir)
    
    query="*"
        
    odir = ds_index_query(index, query, tdir, sources=["_name", "file-info", "description", "main-text"],
                          force=True, limit=1000)        
    
    return odir

def prepare_data(json_files, data_file):

    fw = open(data_file, "w")
    
    for json_file in json_files:

        data=[]
        
        with open(json_file, "r") as fr:
            doc = json.load(fr)

        #_name = doc["_name"]
        if ("description" not in doc) or \
           ("languages" not in doc["description"]) or \
           ("en" not in doc["description"]["languages"]):
            print(f"skipping ...")
            continue
            
        if "file-info" in doc:
            dhash = doc["file-info"]["document-hash"]
        else:
            #print(doc)
            dhash = -1

        """
        if "references" in doc:
            for item in doc["references"]:
                data.append({"label":"reference", "text":item["text"], "document-hash":dhash})

        if "description" in doc:

            desc = doc["description"]

            if "title" in desc:
                #data.append({"label":"title", "text":desc["title"], "document-hash":dhash})
                data.append({"label":"text", "text":desc["title"], "document-hash":dhash})
            
            if "abstract" in desc:
                for item in desc["abstract"]:
                    data.append({"label":"text", "text":item, "document-hash":dhash})

            affiliations=[]
            if "affiliations" in desc:                    
                for item in desc["affiliations"]:
                    affiliations.append(item["name"])
                    #data.append({"label":"affiliation", "text":item["name"], "document-hash":dhash})
                    data.append({"label":"meta-data", "text":item["name"], "document-hash":dhash})

            authors=[]                    
            if "authors" in desc:                                        
                for item in desc["authors"]:
                    authors.append(item["name"])
                    #data.append({"label":"person_name", "text":item["name"], "document-hash":dhash})
                    data.append({"label":"meta-data", "text":item["name"], "document-hash":dhash})

            if len(authors)>1:
                data.append({"label":"meta-data", "text": ", ".join(authors), "document-hash":dhash})
                data.append({"label":"meta-data", "text": "; ".join(authors), "document-hash":dhash})

            if len(affiliations)>1:
                data.append({"label":"meta-data", "text": ", ".join(affiliations), "document-hash":dhash})
                data.append({"label":"meta-data", "text": "; ".join(affiliations), "document-hash":dhash})                
                
            if len(authors)>=1 and len(affiliations)>=1:

                for _ in authors:
                    for __ in affiliations:
                        data.append({"label":"meta-data", "text": " ".join([_, __]), "document-hash":dhash})
        """
        
        if "main-text" in doc:

            N = len(doc["main-text"])

            title_ind=len(doc["main-text"])
            abs_beg=len(doc["main-text"])
            
            ref_beg=len(doc["main-text"])
            ref_end=len(doc["main-text"])
            
            for i,item in enumerate(doc["main-text"]):

                if "text" not in item:
                    continue
                
                label = item["type"].lower()
                text = item["text"].lower().strip()

                if "title" == label:
                    title_ind=i
                
                if ("subtitle" in label) and ("abstract" in text):
                    abs_beg=i

                if ("subtitle" in label) and ("introduction" in text):
                    intro_beg=i                    
                    
                if ("title" in label) and ("reference" in text):
                    ref_beg=i

                if ref_end==N and ref_beg<N and i>ref_beg and ("title" in label) and ("reference" not in text):
                    ref_end=i

            if title_ind==N or abs_beg==N or ref_beg==N:
                print(f"skipping: {dhash}")
                continue
                    
            for i,item in enumerate(doc["main-text"]):

                if "text" not in item:
                    continue

                type_ = item["type"]
                label = item["type"]
                text = item["text"]

                if title_ind<i and i<abs_beg:
                    label = "meta-data"
                elif ref_beg<i and i<ref_end:
                    label = "reference"
                else:
                    label = "text"

                """
                if "title" in label:
                    print(tc.yellow(f"{label}, {type_}: {text[0:48]}"))
                elif "person" in label or "affi" in label:
                    print(tc.green(f"\t{label}, {type_}: {text[0:48]}"))                    
                elif "text" in label:
                    print(f"\t{label}, {type_}: {text[0:48]}")
                elif "caption" in label:
                    print(tc.red(f"\t{label}, {type_}: {text[0:48]}"))
                elif "reference" in label:
                    print(tc.blue(f"\t{label}, {type_}: {text[0:48]}"))
                else:
                    print(tc.red(f"\t{label}, {type_}: {text[0:48]}"))
                """
                
                data.append({"label":label, "text":item["text"], "document-hash":dhash})

        for item in data:
            fw.write(json.dumps(item)+"\n")
        data=[]

    fw.close()
        
def process_data(dfile, afile):

    """
    config = {
        "mode" : "apply",
        "order" : True,
        "models": "numval,link,language"
    }
    """
    
    fr = open(dfile, "r")
    fw = open(afile, "w")
    
    #model = andromeda_nlp.nlp_model()
    #model.initialise(config)

    model = init_nlp_model("numval,link,language")
    
    total=0
    
    while True:

        try:
            line = fr.readline()
            item = json.loads(line)
        except:
            #print(f"could not parse {line}")
            break

        text = item["text"]
        label = item["label"]
        dhash = item["document-hash"]

        if len(text)<8:
            #print(tc.yellow(f"skipping text {label}: {text}"))
            continue
        
        nlpres = model.apply_on_text(item["text"])

        nlpres["document-hash"] = dhash
        
        tind = nlpres["properties"]["headers"].index("type")
        lind = nlpres["properties"]["headers"].index("label")
        cind = nlpres["properties"]["headers"].index("confidence")

        """
        for i,row in enumerate(nlpres["properties"]["data"]):
            if row[tind]=="language" and row[lind]!="en":
                print(f"skipping text of {row[lind]}: {text[0:48]}")
                continue
        """
        
        found=False
        for i,row in enumerate(nlpres["properties"]["data"]):
            if row[tind]=="semantic":
                row[lind] = label
                row[cind] = 1.0
                
                found = True

        if not found:
            nlpres["properties"]["data"].append(["semantic", label, 1.0])
            found = True

        if random.random()<0.9:
            nlpres["training-sample"] = True
        else:
            nlpres["training-sample"] = False

        if found:                    
            fw.write(json.dumps(nlpres)+"\n")
            total += 1
            
    fr.close()
    fw.close()

    print(tc.green(f"found {total} training-items"))
    
# To train a FST model with HPO, one can use
# 
# `./fasttext supervised -input <path-to-train.txt> -output model_name -autotune-validation <<path-to-valid.txt>> -autotune-duration 600 -autotune-modelsize 1M`
#
#  => the parameters can be found via
#
# `./fasttext dump model_cooking.bin args`
#
def train_fst(train_file, model_file, metrics_file):

    #model = andromeda_nlp.nlp_model()
    model = nlp_model()
                
    configs = model.get_train_configs()    
    print(json.dumps(configs, indent=2))

    for config in configs:
        if config["mode"]=="train" and \
           config["model"]=="semantic":

            config["args"] = {}
            config["hpo"]["autotune"] = True
            config["hpo"]["duration"] = 360
            config["hpo"]["modelsize"] = "1M"

            config["args"]["n-gram"] = 0
            
            config["files"]["model-file"] = model_file
            config["files"]["train-file"] = train_file 
            config["files"]["metrics-file"] = metrics_file

            model.train(config)
    
if __name__ == '__main__':

    mode, root_dir = parse_arguments()

    sdir = os.path.join(root_dir, "documents")

    data_file = os.path.join(root_dir, "nlp-train-semantic.data.jsonl")
    annot_file = os.path.join(root_dir, "nlp-train-semantic.annot.jsonl")

    fst_model_file = os.path.join(root_dir, "fst_semantic")
    fst_metrics_file = os.path.join(root_dir, "fst_semantic.metrics.txt")
    
    if mode=="all" or mode=="retrieve":
        tdir = retrieve_data_pubmed(sdir)
        json_files = sorted(glob.glob(os.path.join(tdir, "*.json")))
                
        tdir = retrieve_data(sdir, "acl")
        json_files += sorted(glob.glob(os.path.join(tdir, "*.json")))

        tdir = retrieve_data(sdir, "arxiv")
        json_files += sorted(glob.glob(os.path.join(tdir, "*.json")))        
        
        print(f"results saved in {sdir}: ", len(json_files))
        
    if mode=="all" or mode=="prepare":

        json_files = sorted(glob.glob(os.path.join(sdir, "pubmed/*.json")))
        json_files += sorted(glob.glob(os.path.join(sdir, "acl/*.json")))
        json_files += sorted(glob.glob(os.path.join(sdir, "arxiv/*.json")))

        print("#-files: ", len(json_files))
        
        prepare_data(json_files, data_file)
        process_data(data_file, annot_file)

    if mode=="all" or mode=="train":

        train_fst(annot_file, fst_model_file, fst_metrics_file)    



#!/usr/bin/env python

import os

import json
import glob

import argparse
import textwrap

from tabulate import tabulate

#from ds_convert import convert_pdffile

import pandas as pd
import matplotlib.pyplot as plt

import andromeda_nlp
import andromeda_glm

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'glm_docqa',
        description = 'Do Q&A on pdf document',
        epilog = 'Text at the bottom of help')

    parser.add_argument('--glm-dir', required=True,
                        type=str,
                        help="directory of GLM model")

    parser.add_argument('--qa-pairs', required=False, 
                        type=str, default="prompt",
                        help="CSV file with QA pairs or `prompt`")
    
    parser.add_argument('--models', required=False,                        
                        type=str, default="name;verb;term;abbreviation",
                        help="set NLP models (e.g. `term;sentence`)")
    
    args = parser.parse_args()

    return args.glm_dir, args.qa_pairs, args.models
    
def load_nlp(models:str="name;conn;verb;term;language;reference;abbreviation"):

    nlp_model = andromeda_nlp.nlp_model()

    config = nlp_model.get_apply_configs()[0]
    config["models"] = models
    
    nlp_model.initialise(config)    

    return nlp_model

def load_glm(path:str):

    config = {
        "IO": {
	    "load": {
                "root": path
            }
        }
    }
    
    glm_model = andromeda_glm.glm_model()
    glm_model.load(config)

    return glm_model

def apply_nlp_on_doc(doc_i, nlp_model):

    doc_j = nlp_model.apply_on_doc(doc_i)

def apply_nlp_on_text(text, nlp_model):

    res = nlp_model.apply_on_text(text)    

    return res
    
def analyse_prompt(prompt, nlp_model):

    res = nlp_model.apply_on_text(prompt)

    print(res.keys())
    print(tabulate(res["word-tokens"]["data"], headers=res["word-tokens"]["headers"]))
    print(tabulate(res["entities"]["data"], headers=res["entities"]["headers"]))

    type_ind = res["entities"]["headers"].index("type")
    name_ind = res["entities"]["headers"].index("name")
    
    terms=[]
    for i,row in enumerate(res["entities"]["data"]):
        if(row[type_ind]=="term"):
            terms.append([row[name_ind]])

    return terms

def show_query_result(res, max_nodes=16):

    wrapper = textwrap.TextWrapper(width=50)
    
    print("overview: \n", tabulate(res["overview"]["data"],
                                   headers=res["overview"]["headers"]), "\n")

    for i,item in enumerate(res["result"]):
        
        headers = item["nodes"]["headers"]
        data = item["nodes"]["data"]

        for j,row in enumerate(data):
            text = row[headers.index("text")]
            print("text: ", text)
            
            data[j][headers.index("text")] = "\n".join(wrapper.wrap(text))
                    
        print(f"operation {i}: \n", tabulate(data[0:max_nodes], headers=headers), "\n")

def expand_terms(terms):

    for term in terms:
        print(term)
        
        qry = andromeda_glm.glm_query()
        qry.select({"nodes":[[term]]})
        qry.filter_by({"mode": "node-flavor", "node-flavors":["token"]})
        flid = qry.get_last_flid()
        qry.traverse({"edge":"to-root"})
        qry.traverse({"edge":"from-root"})
        qry.filter_by({"mode": "node-flavor", "node-flavors":["term"]})

        qry.filter_by({"mode": "contains", "contains-flid":flid})        
        qry.traverse({"edge":"to-sent"})
        
        config = qry.to_config()    
        #print("query: ", json.dumps(config, indent=2))    
        
        res = glm_model.query(config)
        if "status" in res and res["status"]=="success":
            show_query_result(res)
        else:
            print(res)
            #print(res["status"], ": ", res["message"])
            
def do_qa(nlp_model, glm_model):

    while True:
        
        prompt = input("question: ")
        #prompt = "What is the income of IBM in 2022?"
        #prompt = "net-zero"
        if(prompt=="q"):
            break
        
        #terms = analyse_prompt(prompt, nlp_model)
        terms = prompt.split(" ")
        
        expand_terms(terms)
        
        """
        qry = andromeda_glm.glm_query()
        qry.select({"nodes":search_terms})
        qry.filter_by({"mode": "node-flavor", "node-flavors":["term"]})
        qry.traverse({"edge":"to-table"})
        
        config = qry.to_config()    
        print("query: ", json.dumps(config, indent=2))    

        res = glm_model.query(config)
        show_query_result(res)
        """

        #break

def compute_topk_on_documents(df, nlp_model, glm_model):

    topk = {0:0}
    for ind in range(1, 10):
        topk[ind] = 0
        
    for i,row in df.iterrows():
        print(i, row["question"])

        doc_hash = row["doc_hash"]
        question = row["question"]

        context = row["text"]
        
        qres = apply_nlp_on_text(question, nlp_model)
        cres = apply_nlp_on_text(context, nlp_model)
        
        """
        print(json.dumps(res, indent=2))
        print(tabulate(res["instances"]["data"],
                       headers=res["instances"]["headers"]))
        """

        data = cres["instances"]["data"]
        headers = cres["instances"]["headers"]

        insts=[]
        for j,row in enumerate(data):
            insts.append([row[headers.index("type")],
                          row[headers.index("subtype")],
                          row[headers.index("name")]])

        """
        print(f"context: {context}\n")
            
        print("instances: ")
        print(tabulate(insts, headers=["type", "subtype", "name"]), "\n")
        """
        
        data = qres["instances"]["data"]
        headers = qres["instances"]["headers"]
        
        insts=[]
        for j,row in enumerate(data):
            insts.append([row[headers.index("type")],
                          row[headers.index("subtype")],
                          row[headers.index("name")]])
        """
        print(f"question: {question}\n")
            
        print("instances: ")
        print(tabulate(insts, headers=["type", "subtype", "name"]), "\n")
        """
        
        terms=[]
        for j,row in enumerate(insts):
            if "term"==row[0]:
                term=row[2].split()
                terms.append(term)
                
        qry = andromeda_glm.glm_query()
        qry.select({"nodes":terms})
        qry.traverse({"edge":"to-doc"})

        config = qry.to_config()    
        #print("query: ", json.dumps(config, indent=2))            

        out = glm_model.query(config)
        #print(json.dumps(out, indent=2))
                
        if out["status"]=="success":

            docs = pd.DataFrame(out["result"][1]["nodes"]["data"],
                                columns=out["result"][1]["nodes"]["headers"])

            doc_hashes = list(docs["text"])
            
            for k,v in topk.items():
                if k==0:
                    topk[k] += 1
                elif doc_hash in doc_hashes[0:k]:
                    topk[k] += 1

    print(json.dumps(topk, indent=2))

    x=[]
    y=[]
    for i in range(1,10):
        x.append(i)
        y.append(topk[i]/topk[0])
    
    plt.figure(1)
    plt.plot(x,y, "r.-", label="doc-topk")
    plt.ylim(0,1.05)
    plt.legend(loc="lower right")
    plt.show()

def compute_topk_on_element(df, nlp_model, glm_model):

    topk = {0:0}
    for ind in range(1, 10):
        topk[ind] = 0
        
    for i,row in df.iterrows():
        print(i, row["question"])

        doc_hash = row["doc_hash"]
        question = row["question"]

        context = row["text"]
        
        qres = apply_nlp_on_text(question, nlp_model)
        cres = apply_nlp_on_text(context, nlp_model)    

        data = qres["instances"]["data"]
        headers = qres["instances"]["headers"]
        
        insts=[]
        for j,row in enumerate(data):
            insts.append([row[headers.index("type")],
                          row[headers.index("subtype")],
                          row[headers.index("name")]])
        """
        print(f"question: {question}\n")
            
        print("instances: ")
        print(tabulate(insts, headers=["type", "subtype", "name"]), "\n")
        """
        
        terms=[]
        for j,row in enumerate(insts):
            if "term"==row[0]:
                term=row[2].split()
                terms.append(term)
                
        qry = andromeda_glm.glm_query()
        qry.select({"nodes":terms})
        qry.traverse({"edge":"to-text"})

        config = qry.to_config()    
        #print("query: ", json.dumps(config, indent=2))            

        out = glm_model.query(config)
        #print(json.dumps(out, indent=2))
                
        if out["status"]=="success":

            docs = pd.DataFrame(out["result"][1]["nodes"]["data"],
                                columns=out["result"][1]["nodes"]["headers"])

            doc_hashes = list(docs["text"])
            
            for k,v in topk.items():
                if k==0:
                    topk[k] += 1
                elif doc_hash in doc_hashes[0:k]:
                    topk[k] += 1

    print(json.dumps(topk, indent=2))

    
if __name__ == '__main__':

    glm_dir, qa_pairs_file, models = parse_arguments()

    glm_model = load_glm(glm_dir)    
    nlp_model = load_nlp(models)    

    df = pd.read_csv(qa_pairs_file)

    compute_topk_on_documents(df, nlp_model, glm_model)
    #compute_topk_on_element(df, nlp_model, glm_model)



#!/usr/bin/env python

import os

import json
import glob

import datetime
import argparse

import pandas as pd

from ds_utils import create_docs_dir, ds_list_indices, ds_index_query

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'ds_query',
        description = 'Query and download documents on the Deep Search platform',
        epilog =
"""
examples of execution: 

1. search for documents:

    poetry run python ./deepsearch_glm/ds_query.py --index esg-reports --query "\\\"net zero\\\""
    poetry run python ./deepsearch_glm/ds_query.py --index patent-uspto --query "\\\"global warming potential\\\" AND \\\"etching\\\""
    poetry run python ./deepsearch_glm/ds_query.py --index arxiv --query "\\\"quantum computing\\\""


""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--index', required=True,
                        type=str,
                        help="Deep Search document index")

    parser.add_argument('--query', required=True,
                        type=str, 
                        help="Query for document")

    parser.add_argument('--output-dir', required=False,
                        type=bool, default=create_docs_dir(),
                        help="output directory for documents")        

    args = parser.parse_args()

    return args.index, args.query, args.output_dir

if __name__ == '__main__':

    index, query, odir = parse_arguments()

    indices = ds_list_indices()

    found=False
    for item in indices:
        if item["Index"]==index:
            found=True

    if not found:
        print(pd.DataFrame(indices))
        print(f"index `{index}` not found")
        exit(-1)

    else:
        odir = ds_index_query(index, query, odir, force=True)        
        print(f"files downloaded: {odir}")



#!/usr/bin/env python

import os

import json
import glob

import argparse
import textwrap
import datetime

from tabulate import tabulate

from deepsearch_glm.glm_utils import create_glm_dir, create_glm_from_docs

#import andromeda_glm

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'create_glm_from_docs',
        description = 'Create GLM from Deep Search documents',
        epilog =
"""
examples of execution: 

1.a run on single document (pdf or json) with default NLP models (=`term`):
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --pdf './data/documents/articles/2305.02334.pdf'
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --json './data/documents/articles/2305.02334.json'

2. run on multiple documents:
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --pdf './data/documents/articles/*.pdf'
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --json './data/documents/articles/*.json'

3. run on multiple documents with non-default models:
    poetry run python ./deepsearch_glm/create_glm_from_docs.py --pdf './data/documents/articles/2305.*.pdf' --models 'language;term;abbreviation'

""",
        formatter_class=argparse.RawTextHelpFormatter)
        
    parser.add_argument('--pdf', required=False,
                        type=str, default=None,
                        help="filename(s) of pdf document")
    
    parser.add_argument('--json', required=False,
                        type=str, default=None,
                        help="filename(s) of json document")

    parser.add_argument('--models', required=False,
                        type=str, default="name;verb;term;abbreviation",
                        help="set NLP models (e.g. `term;sentence`)")
    
    parser.add_argument('--force-convert', required=False,
                        type=bool, default=False,
                        help="force pdf conversion")

    parser.add_argument('--output-dir', required=False,
                        type=str, default=create_glm_dir(),
                        help="output root directory for GLM")    

    args = parser.parse_args()

    pdf = args.pdf
    json = args.json

    if pdf==None and json==None:
        exit(-1)
        
    if pdf!=None:
        pdf_files=sorted(glob.glob(pdf))
    else:
        pdf_files=[]
        
    if json!=None:
        json_files=sorted(glob.glob(json))
    else:
        json_files=[]
        
    return pdf_files, json_files, args.models, args.force_convert, args.output_dir
    
if __name__ == '__main__':

    pdf_files, json_files, model_names, force_convert, odir = parse_arguments()

    if len(pdf_files)>0:
        new_json_files = convert_pdffiles(pdf_files, force=force_convert)

        for _ in new_json_files:
            json_files.append(_)

    json_files = sorted(list(set(json_files)))        

    #odir = create_glm_dir()
    
    glm = create_glm_from_docs(odir, json_files, model_names)
    print(f" --> GLM saved to: {odir}")

    



#!/usr/bin/env python

import os
import json
import glob

import argparse
"""
import textwrap
wrapper = textwrap.TextWrapper(width=70)

import pandas as pd
pd.options.display.width=48

import textColor as tc


"""

import andromeda_nlp

from deepsearch_glm.nlp_utils import init_nlp_model, print_on_shell

default_text = """A team of physicists at Université Paris-Saclay has, for the first time, observed spontaneous quasi-crystal self-assembly. The observation occurred during an experiment they were conducting with tiny vibrating magnetic spheres. The team has written a paper describing their experiment and have posted it on the arXiv preprint server while they await peer review."""

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'nlp_apply_on_text',
        description = 'Apply NLP on text',
        epilog =
"""
examples of execution:

1. apply NLP on default text with default models (=`langauge`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py

2. apply NLP on a single piece of text with default model (=`langauge`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py --text "FeSe is a superconductor with Tc of 30 K."

3. apply NLP on default text with specific models (=`verb;term;conn;semantic;abbreviation`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py --model-names "verb;term;conn;semantic;abbreviation"

4. apply NLP on text from prompt with specific models (=`verb;term`):

    poetry run python ./deepsearch_glm/apply_nlp_on_text.py --model-names "verb;term" --interactive True

""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-t', '--text', required=False,
                        type=str, default=default_text,
                        help="text on which the NLP is applied.")

    parser.add_argument('-m', '--model-names', required=False,
                        type=str, default="verb;term;conn;semantic;abbreviation",
                        help="model_names to be applied")

    parser.add_argument('--interactive', required=False,
                        type=bool, default=False,
                        help="run on text interactively")
    
    args = parser.parse_args()
    
    return args.text, args.model_names, args.interactive

if __name__ == '__main__':

    text, model_names, interactive = parse_arguments()
    
    model = andromeda_nlp.nlp_model()
    print("resource-dir: ", model.get_resources_path())

    model.initialise_models(model_names)

    if interactive:

        while True:
            text = input("text: ")

            if text in ["q", "quit"]:
                break

            result = model.apply_on_text(text)
            print_on_shell(text, result)

    else:

        result = model.apply_on_text(text)
        print_on_shell(text, result)




#!/usr/bin/env python

import os
import copy
import json
import glob

import argparse
import textwrap

import numpy as np
import pandas as pd

from tabulate import tabulate
from PIL import Image, ImageDraw

#import andromeda_nlp
from deepsearch_glm.andromeda_nlp import nlp_model

def parse_arguments():

    parser = argparse.ArgumentParser(
        prog = 'nlp_analyse_docs',
        description = 'Analyse NLP on `Deep Search` documents ()',
        epilog =
"""
examples of execution:

1.a run on single document (pdf or json) with default model (=`langauge`):

    poetry run python ./deepsearch_glm/nlp_analyse_docs.py --json ./data/documents/articles/2305.02334.nlp.json

""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--json', required=True,
                        type=str, default=None,
                        help="filename(s) of json document")
    
    parser.add_argument('--mode', required=False,
                        type=str, default="all",
                        help="selection of what to show (default=all, example='page=0;refereces')")

    args = parser.parse_args()
    
    json_files=sorted(glob.glob(args.json))

    return json_files, args.mode

def extract_meta(doc):
    return

def show_page(doc, page_num=1, show_orig=True):

    print(doc.keys())
    
    page_dims = doc["page-dimensions"]
    df_dims = pd.read_json(json.dumps(page_dims), orient='records')
    print(df_dims)
    
    page_items = doc["page-elements"]
    print(page_items[0])

    # FIXME: ineffecient but works ...
    df = pd.read_json(json.dumps(page_items), orient='records')
    print(df)

    df_page = df[df["page"]==page_num]

    print(df_dims[df_dims["page"]==page_num])

    ph = int(df_dims[df_dims["page"]==page_num]["height"])
    pw = int(df_dims[df_dims["page"]==page_num]["width"])
    
    #ph = int(df_dims[df_dims["page"]==page_num]["height"][0])
    #pw = int(df_dims[df_dims["page"]==page_num]["width"][0])

    print("height: ", ph)
    print("width: ", pw)
    
    df_page["bbox"] = df_page["bbox"].apply(lambda bbox: [bbox[0], ph-bbox[3], bbox[2], ph-bbox[1]])
    #df_page = df_page.apply(lambda row: [row["bbox"][0], ph-row["bbox"][3], row["bbox"][2], ph-row["bbox"][1]], axis=1)
    
    factor = 2 if show_orig else 1 
    
    # Create a new image with a white background
    image = Image.new("RGB", (factor*pw, ph), "white")

    # Create an ImageDraw object to draw on the image
    orig_order=[]
    points=[]

    draw = ImageDraw.Draw(image)
    for i,row in df_page.iterrows():

        bbox = row["bbox"]
        draw.rectangle(bbox, outline="red", width=2)

        orig_order.append(row["orig-order"])
        points.append((int((bbox[0]+bbox[2])/2.0),
                       int((bbox[1]+bbox[3])/2.0)))

    draw.line(points, fill="blue", width=2)

    for i,point in enumerate(points):
        draw.text(point, str(i), fill="black")
    
    if show_orig:
        for i,row in df_page.iterrows():
            bbox = row["bbox"]
            bbox[0] += pw
            bbox[2] += pw
            draw.rectangle(bbox, outline="red", width=2)

        opoints = copy.deepcopy(points)
        for i,j in enumerate(orig_order):
            #print(i, "\t", j, "\t", len(points), "\t", len(opoints))
            x,y = points[i]
            opoints[j] = (x+pw,y)
        
        draw.line(opoints, fill="orange", width=2)    

        for i,point in enumerate(opoints):
            draw.text(point, str(i), fill="black")
        
    image.show()        
    
def extract_text(doc):

    page_items = doc["page-elements"]
    texts = doc["texts"]

    print("\n\t TEXT: \n")
    
    wrapper = textwrap.TextWrapper(width=70)
    
    for item in doc["texts"]:

        labels=[]
        for row in item["properties"]["data"]:
            labels.append(row[item["properties"]["headers"].index("label")])
        
        type_ = item["type"]
        print(f"text: {type_}, labels: ", ", ".join(labels))
        for line in wrapper.wrap(text=item["text"]):
            print(f"\t{line}")

        print("")
    return

def extract_sentences(doc):

    df = pd.DataFrame(doc["instances"]["data"],
                      columns=doc["instances"]["headers"])

    entity_types = df["type"].value_counts()
    print(entity_types)
    
    sents = df[df["type"]=="sentence"]
    print(sents)
    
    return

def extract_tables(doc):
    
    wrapper = wrapper = textwrap.TextWrapper(width=70)

    print("\n\t TABLES: \n")
    
    for i,item in enumerate(doc["tables"]):

        print(f"#/tables/{i}: ", len(item["captions"]))
        if len(item["captions"])>0:
            for line in wrapper.wrap(text=item["captions"][0]["text"]):
                print(f"\t{line}")
    
    return

def extract_figures(doc):

    print("\n\t FIGURES: \n")
    
    wrapper = wrapper = textwrap.TextWrapper(width=70)
    
    for i,item in enumerate(doc["figures"]):

        print(f"#/figures/{i}", len(item["captions"]))
        if len(item["captions"])>0:
            for line in wrapper.wrap(text=item["captions"][0]["text"]):
                print(f"\t{line}")
    
    return

def extract_references(doc):

    print("\n\t REFERENCES: \n")
    
    page_items = doc["page-elements"]
    texts = doc["texts"]

    df = pd.DataFrame(doc["instances"]["data"],
                      columns=doc["instances"]["headers"])
    
    wrapper = wrapper = textwrap.TextWrapper(width=70)
    
    for i,item in enumerate(doc["texts"]):

        path = f"#/texts/{i}"
        
        labels=[]
        for row in item["properties"]["data"]:
            labels.append(row[item["properties"]["headers"].index("label")])

        if "reference" in labels:

            print(f"text: ") #{type_}, labels: ", ",".join(labels))
            for line in wrapper.wrap(text=item["text"]):
                print(f"\t{line}")            
                
            refs = df[df["subj_path"]==path][["subj_path", "type", "subtype", "original"]]
            print(refs)            
                
    return

if __name__ == '__main__':

    json_files, mode = parse_arguments()

    for json_file in json_files:

        print(f" --> reading {json_file}")
        with open(json_file, "r") as fr:
            doc = json.load(fr)

        if "page" in mode or mode=="all":
            show_page(doc, page_num=1)

        if "text" in mode or mode=="all":
            extract_text(doc)

        if "sentence" in mode or mode=="all":            
            extract_sentences(doc)

        if "table" in mode or mode=="all":            
            extract_tables(doc)

        if "figure" in mode or mode=="all":            
            extract_figures(doc)

        if "reference" in mode or mode=="all":            
            extract_references(doc)
